{****************************************************************************}
{                      Adapted for Data Master 2000                          }
{    original code from Borland Turbo Pascal Numerical Methods Toolbox       }
{****************************************************************************}
unit LeastSqr;

interface

const
  TNNearlyZero = 1E-015; {???}

  TNRowSize = 32;  { Maximum number of terms in Least Squares fit -
  cannot be too long because of errors, so remains constant! }

type
  Float = Double; { 8 byte real, requires 8087 math chip }
  TNColumnVector = array of Float; // dynamic array !
  TNRowVector = array[1..TNRowSize] of Float;
  TNMatrix = array of TNRowVector; // two-dimensional dynamic array !!!
  TNSquareMatrix =array[1..TNRowSize] of TNRowVector;

  TFitFunction=class  // we will use class methods instead of proc. variables
    public
    class procedure Transform(NumPoints  : integer;
                          var XData      : TNColumnVector;
                          var YData      : TNColumnVector;
                          var Multiplier : Float;
                          var Constant   : Float;
                          var WData      : TNColumnVector;
                          var ZData      : TNColumnVector;
                          var Error      : byte); virtual; abstract;
    class procedure InverseTransform(Multiplier : Float;
                                     Constant   : Float;
                                     var YFit   : Float); virtual; abstract;
    class procedure CreateBasisFunctions(NumPoints : integer;
                                     var NumTerms  : integer;
                                         var WData : TNColumnVector;
                                         var Basis : TNmatrix); virtual; abstract;
    class procedure TransformSolution(NumTerms    : integer;
                                  var OldSolution : TNRowVector;
                                      Multiplier  : Float;
                                      Constant    : Float;
                                  var NewSolution : TNRowVector); virtual; abstract;
  end;

  TFitFunctionClass=class of TFitFunction;

  TPolyFitFunction=class(TFitFunction)
    public
    class procedure Transform(NumPoints  : integer;
                          var XData      : TNColumnVector;
                          var YData      : TNColumnVector;
                          var Multiplier : Float;
                          var Constant   : Float;
                          var WData      : TNColumnVector;
                          var ZData      : TNColumnVector;
                          var Error      : byte); override;
    class procedure InverseTransform(Multiplier : Float;
                                     Constant   : Float;
                                     var YFit   : Float); override;
    class procedure CreateBasisFunctions(NumPoints : integer;
                                     var NumTerms  : integer;
                                         var WData : TNColumnVector;
                                         var Basis : TNmatrix); override;
    class procedure TransformSolution(NumTerms    : integer;
                                  var OldSolution : TNRowVector;
                                      Multiplier  : Float;
                                      Constant    : Float;
                                  var NewSolution : TNRowVector); override;
  end;

  TFourierFitFunction=class(TFitFunction)
    public
    class procedure Transform(NumPoints  : integer;
                          var XData      : TNColumnVector;
                          var YData      : TNColumnVector;
                          var Multiplier : Float;
                          var Constant   : Float;
                          var WData      : TNColumnVector;
                          var ZData      : TNColumnVector;
                          var Error      : byte); override;
    class procedure InverseTransform(Multiplier : Float;
                                     Constant   : Float;
                                     var YFit   : Float); override;
    class procedure CreateBasisFunctions(NumPoints : integer;
                                     var NumTerms  : integer;
                                         var WData : TNColumnVector;
                                         var Basis : TNmatrix); override;
    class procedure TransformSolution(NumTerms    : integer;
                                  var OldSolution : TNRowVector;
                                      Multiplier  : Float;
                                      Constant    : Float;
                                  var NewSolution : TNRowVector); override;
  end;

  TPowerFitFunction=class(TFitFunction)
    public
    class procedure Transform(NumPoints  : integer;
                          var XData      : TNColumnVector;
                          var YData      : TNColumnVector;
                          var Multiplier : Float;
                          var Constant   : Float;
                          var WData      : TNColumnVector;
                          var ZData      : TNColumnVector;
                          var Error      : byte); override;
    class procedure InverseTransform(Multiplier : Float;
                                     Constant   : Float;
                                     var YFit   : Float); override;
    class procedure CreateBasisFunctions(NumPoints : integer;
                                     var NumTerms  : integer;
                                         var WData : TNColumnVector;
                                         var Basis : TNmatrix); override;
    class procedure TransformSolution(NumTerms    : integer;
                                  var OldSolution : TNRowVector;
                                      Multiplier  : Float;
                                      Constant    : Float;
                                  var NewSolution : TNRowVector); override;
  end;

  TExpoFitFunction=class(TFitFunction)
    public
    class procedure Transform(NumPoints  : integer;
                          var XData      : TNColumnVector;
                          var YData      : TNColumnVector;
                          var Multiplier : Float;
                          var Constant   : Float;
                          var WData      : TNColumnVector;
                          var ZData      : TNColumnVector;
                          var Error      : byte); override;
    class procedure InverseTransform(Multiplier : Float;
                                     Constant   : Float;
                                     var YFit   : Float); override;
    class procedure CreateBasisFunctions(NumPoints : integer;
                                     var NumTerms  : integer;
                                         var WData : TNColumnVector;
                                         var Basis : TNmatrix); override;
    class procedure TransformSolution(NumTerms    : integer;
                                  var OldSolution : TNRowVector;
                                      Multiplier  : Float;
                                      Constant    : Float;
                                  var NewSolution : TNRowVector); override;
  end;

  TLorentzFitFunction=class(TPolyFitFunction)
    public
    class procedure Transform(NumPoints: integer;
      var XData,YData: TNColumnVector; var Multiplier,Constant: Float;
      var WData,ZData: TNColumnVector; var Error: byte); override;
    class procedure CreateBasisFunctions(NumPoints: integer;
      var NumTerms: integer; var WData: TNColumnVector;
      var Basis: TNmatrix); override;
    class procedure TransformSolution(NumTerms: integer;
      var OldSolution: TNRowVector; Multiplier,Constant: Float;
      var NewSolution: TNRowVector); override;
  end;

  TSineFitFunction=class(TFourierFitFunction)
    public
    class procedure CreateBasisFunctions(NumPoints: integer;
      var NumTerms: integer; var WData: TNColumnVector;
      var Basis: TNmatrix); override;
    class procedure TransformSolution(NumTerms: integer;
      var OldSolution: TNRowVector; Multiplier,Constant: Float;
      var NewSolution: TNRowVector); override;
  end;

procedure LeastSquares(NumPoints         : integer;
                   var XData             : TNColumnVector;
                   var YData             : TNColumnVector;
                       NumTerms          : integer;
                   var Solution          : TNRowVector;
                   var YFit              : TNColumnVector;
                   var Residuals         : TNColumnVector;
                   var StandardDeviation : Float;
                   var Variance          : Float;
                   var Error             : byte;
                       Fit               : TFitFunctionClass); // !!!

{----------------------------------------------------------------------------}
{-                                                                          -}
{-   Input: NumPoints, XData, YData, NumTerms                               -}
{-   Output: Solution, YFit, Residuals, StandardDeviation, Error            -}
{-                                                                          -}
{-   Purpose:  Given NumPoints data points of the form (X, Y), this         -}
{-             procedure finds the least square solution of NumTerms terms  -}
{-             (NumTerms <= NumPoints) to then matrix equation AC = B       -}
{-             where A is a NumPoints by NumTerms matrix, B is a            -}
{-             NumPoints vector and C is the least squares solution.  The   -}
{-             elements of A are A[i, j] = Tj(X[i]) where Tj is the jth     -}
{-             basis vector and X[i] is the X-value of the ith data point.  -}
{-             The basis vectors are created by a separate include file,    -}
{-             or module.  The choice of module will determine whether      -}
{-             the least squares solution is a polynomial fit, trigono-     -}
{-             metric fit (Fourier series), power fit (e.g. Y=ax^b, where   -}
{-             b is fractional), exponential fit (e.g. Y=a-Exp(bx)), or     -}
{-             logarithmic fit (e.g. Y=a-Ln(bx)).  The user may also        -}
{-             create modules for other functional forms.  See the          -}
{-             documentation for details.                                   -}
{-                                                                          -}
{- User-Defined Types: TNColumnVector = array[1..TNColumnSize] of real;     -}
{-                     TNRowVector = array[1..TNRowSize] of real;           -}
{-                                                                          -}
{-     Global Variables: NumPoints         : integer; Number of data points -}
{-                       XData             : TNColumnVector; X-value data   -}
{-                       YData             : TNColumnVector; Y-value data   -}
{-                       NumTerms          : integer; Number of terms in    -}
{-                                                    least squares fit     -}
{-                       Solution          : TNRowVector; Least squares     -}
{-                                                        solution in the   -}
{-                                                        given basis       -}
{-                       YFit              : TNColumnVector; Y-values       -}
{-                                                           predicted by   -}
{-                                                          the LS solution -}
{-                       Residuals         :TNColumnVector Difference       -}
{-                                                      between predicted   -}
{-                                                      and actual Y values -}
{-                       StandardDeviation : real; Root of variance         -}
{-                       Error             : byte; Indicates an error       -}
{-                                                                          -}
{-               Errors: 0: No errors                                       -}
{-                       1: NumPoints < 2                                   -}
{-                       2: NumTerms < 1                                    -}
{-                       3: NumTerms > NumPoints                            -}
{-                       4: solution not possible (exact reason             -}
{-                          will depend upon the particular basis)          -}
{-                                                                          -}
{----------------------------------------------------------------------------}

implementation

procedure LeastSquares(NumPoints        : integer;
                   var XData             : TNColumnVector;
                   var YData             : TNColumnVector;
                       NumTerms          : integer;
                   var Solution          : TNRowVector;
                   var YFit              : TNColumnVector;
                   var Residuals         : TNColumnVector;
                   var StandardDeviation : Float;
                   var Variance          : Float;
                   var Error             : byte;
                       Fit               : TFitFunctionClass);

var
  WData : TNColumnVector;          { Transformed X-values  }
  ZData : TNColumnVector;          { Transformed Y-values  }
  Basis : TNmatrix;                { Matrix of basis functions  }

  Multiplier : Float;              { Multiplier and Constant are used in }
  Constant : Float;                { some modules to pass information    }
                                   { from Transform to InverseTransform  }
                                   { or TransformSolution.  These must   }
                                   { therefore have dummy parameters     }
                                   { when Multiplier and Constant are    }
                                   { not used.                           }

procedure InitializeAndFormBasisVectors(NumPoints  : integer;
                                    var XData      : TNColumnVector;
                                    var YData      : TNColumnVector;
                                    var Multiplier : Float;
                                    var Constant   : Float;
                                    var WData      : TNColumnVector;
                                    var ZData      : TNColumnVector;
                                    var NumTerms   : integer;
                                    var Solution   : TNRowVector;
                                    var Basis      : TNmatrix;
                                    var Error      : byte);

{-------------------------------------------------------}
{- Input: NumPoints, XData, NumTerms                   -}
{- Output: Solution, Error                             -}
{-                                                     -}
{- This procedure initializes Solution and Error       -}
{- to zero.  It also checks the data for errors.       -}
{-------------------------------------------------------}
begin
  Error := 0;
  if NumPoints < 2 then
    Error := 1;   { Less than 2 data points  }
  if NumTerms < 1 then
    Error := 2;   { Less than 1 coefficient in the fit  }
  if NumTerms > NumPoints then
    Error := 3;   { Number of data points less than        }
                  { number of terms in Least Squares fit!  }
  FillChar(Solution, SizeOf(Solution), 0);
  if Error = 0 then
  begin
    { The next two procedures are particular to each        }
    { basis.  Consequently, they are included in each       }
    { module, not in this include file.                     }

    { The Transform procedure transforms the input data to  }
    { fit the particular basis.  This may mean taking the   }
    { logarithm, or linearly tranforming the data to a      }
    { particular interval. XData is transformed to WData    }
    { and YData is transformed to ZData.  For some of the   }
    { modules, Multiplier and Constant are used to pass     }
    { information, for others they are dummy variables.     }
    { See the code listing of the appropriate module for    }
    { more information.                                     }
    Fit.Transform(NumPoints, XData, YData, Multiplier,
                              Constant, WData, ZData, Error);

    if Error = 0 then

      { The CreateBasis procedure creates the matrix of       }
      { basis vectors, Basis.  The elements of Basis are:     }
      { Basis[i, j] = Tj(w[i]) where Tj is the jth basis      }
      { and w[i] is the ith data element of WData.            }
      Fit.CreateBasisFunctions(NumPoints, NumTerms, WData, Basis);
  end;
end; { procedure InitializeAndFormBasisVectors }

procedure CreateAndSolveEquations(NumPoints : integer;
                                  NumTerms  : integer;
                              var Basis     : TNmatrix;
                              var ZData     : TNColumnVector;
                              var Solution  : TNRowVector;
                              var Error     : byte);


var
  Coefficients : TNSquareMatrix;
  Constants : TNRowVector;

{------------------------------------------------------------}
{- Input: NumPoints, NumTerms, Basis, ZData                 -}
{- Output: Solution, Error                                  -}
{- This procedure computes and solves the normal equations. -}
{- The normal equations are represented in matrix notation  -}
{- as      Coefficients - Solution = Constants              -}
{- This matrix equation is solved by Gaussian Elimination   -}
{- with partial pivoting (TNToolbox routine: PARTPIVT.INC). -}
{- If no solution exists, Error 3 is returned.              -}
{------------------------------------------------------------}

procedure ComputeNormalEquations(NumPoints, NumTerms : integer;
                             var Basis               : TNmatrix;
                             var YData               : TNColumnVector;
                             var Coefficients        : TNSquareMatrix;
                             var Constants           : TNRowVector);

{---------------------------------------------------------}
{- Input: NumPoints, NumTerms, Basis, YData              -}
{- Output: Coefficients, Constants                       -}
{-                                                       -}
{- This procedure calculates the normal equations.  The  -}
{- normal equations are of the form Ax=b, where A is the -}
{- Coefficients matrix, b is the Constants vector and X  -}
{- is the least squares solution to our problem in the   -}
{- given basis.                                          -}
{- The normal equations are derived from the basis       -}
{- functions and the condition of least squares.  The    -}
{- algorithm to create them is:                          -}
{-      Coefficients[i, j] = Sum from k=1 to NumPoints   -}
{-                            of Basis[k, i]-Basis[k, j] -}
{-                                                       -}
{-     Constants[i] = Sum from k=1 to NumPoints          -}
{-                     YData[k] - Basis[k, i]            -}
{---------------------------------------------------------}

var
  Row, Column, Index : integer;
  Sum : Float;

begin
  for Column := 1 to NumTerms do
  begin
    Sum := 0;
    for Index := 1 to NumPoints do
    Sum := Sum + YData[Index] * Basis[Index, Column];
    Constants[Column] := Sum;
    for Row := Column to NumTerms do
    begin
      Sum := 0;
      for Index := 1 to NumPoints do
        Sum := Sum + Basis[Index, Row] * Basis[Index, Column];
      Coefficients[Row, Column] := Sum;
      Coefficients[Column, Row] := Sum;
    end;
  end;
end; { procedure ComputeNormalEquations }

{---------------------------------------------------------------------------}

procedure Partial_Pivoting(Dimen        : integer;
                           Coefficients : TNSquareMatrix;
                           Constants    : TNRowVector;
                       var Solution     : TNRowVector;
                       var Error        : byte);

{--------------------------------------------------------------------------}
{-                                                                        -}
{-              Input: Dimen, Coefficients, Constants                     -}
{-             Output: Solution, Error                                    -}
{-                                                                        -}
{-           Purpose : calculate the solution of a linear set of          -}
{-                     equations using Gaussian elimination, maximal      -}
{-                     pivoting and backwards substitution.               -}
{-                                                                        -}
{-  User-defined Types : TNRowVector = array[1..TNArraySize] of real;     -}
{-                 TNSquareMatrix = array[1..TNArraySize] of TNRowVector  -}
{-                                                                        -}
{-  Global Variables : Dimen        : integer;     Dimension of the       -}
{-                                                 square matrix          -}
{-                     Coefficients : TNSquareMatrix; Square matrix       -}
{-                     Constants    : TNRowVector; Constants of           -}
{-                                                 each equation          -}
{-                     Solution     : TNRowVector; Unique solution to     -}
{-                                                 the set of equations   -}
{-                     Error        : integer;     Flags if something     -}
{-                                                 goes wrong.            -}
{-                                                                        -}
{-            Errors : 0: No errors;                                      -}
{-                     1: Dimen < 2                                       -}
{-                     2: no solution exists                              -}
{-                                                                        -}
{--------------------------------------------------------------------------}

procedure Initial(Dimen        : integer;
              var Coefficients : TNSquareMatrix;
              var Constants    : TNRowVector;
              var Solution     : TNRowVector;
              var Error        : byte);

{----------------------------------------------------------}
{- Input: Dimen, Coefficients, Constants                  -}
{- Output: Solution, Error                                -}
{-                                                        -}
{- This procedure test for errors in the value of Dimen.  -}
{- This procedure also finds the solution for the         -}
{- trivial case Dimen = 1.                                -}
{----------------------------------------------------------}

begin
  Error := 0;
  if Dimen < 1 then
    Error := 1
  else
    if Dimen = 1 then
      if ABS(Coefficients[1, 1]) < TNNearlyZero then
        Error := 2
      else
        Solution[1] := Constants[1] / Coefficients[1, 1];
end; { procedure Initial }

procedure EROswitch(var Row1 : TNRowVector;
                    var Row2 : TNRowVector);

{-------------------------------------------------}
{- Input: Row1, Row2                             -}
{- Output: Row1, Row2                            -}
{-                                               -}
{- Elementary row operation - switching two rows -}
{-------------------------------------------------}

var
  DummyRow : TNRowVector;

begin
  DummyRow := Row1;
  Row1 := Row2;
  Row2 := DummyRow;
end; { procedure EROswitch }

procedure EROmultAdd(Multiplier   : Float;
                     Dimen        : integer;
                 var ReferenceRow : TNRowVector;
                 var ChangingRow  : TNRowVector);

{-----------------------------------------------------------}
{- Input: Multiplier, Dimen, ReferenceRow, ChangingRow     -}
{- Output: ChangingRow                                     -}
{-                                                         -}
{- row operation - adding a multiple of one row to another -}
{-----------------------------------------------------------}

var
  Term : integer;

begin
  for Term := 1 to Dimen do
    ChangingRow[Term] := ChangingRow[Term] + Multiplier * ReferenceRow[Term];
end; { procedure EROmult&Add }

procedure UpperTriangular(Dimen        : integer;
                      var Coefficients : TNSquareMatrix;
                      var Constants    : TNRowVector;
                      var Error        : byte);

{-----------------------------------------------------------------}
{- Input: Dimen, Coefficients, Constants                         -}
{- Output: Coefficients, Constants, Error                        -}
{-                                                               -}
{- This procedure makes the coefficient matrix upper triangular. -}
{- The operations which perform this are also performed on the   -}
{- Constants vector.                                             -}
{- If one of the main diagonal elements of the upper triangular  -}
{- matrix is zero, then the Coefficients matrix is singular and  -}
{- no solution exists (Error = 2 is returned).                   -}
{-----------------------------------------------------------------}

var
  Multiplier : Float;
  Row, ReferenceRow : integer;

procedure Pivot(Dimen        : integer;
                ReferenceRow : integer;
            var Coefficients : TNSquareMatrix;
            var Constants    : TNRowVector;
            var Error        : byte);

{----------------------------------------------------------------}
{- Input: Dimen, ReferenceRow, Coefficients                     -}
{- Output: Coefficients, Constants, Error                       -}
{-                                                              -}
{- This procedure searches the ReferenceRow column of the       -}
{- Coefficients matrix for the largest non-zero element below   -}
{- the diagonal. If it finds one, then the procedure switches   -}
{- rows so that the largest non-zero element is on the          -}
{- diagonal. It also switches the corresponding elements in     -}
{- the Constants vector. If it doesn't find a non-zero element, -}
{- the matrix is singular and no solution exists                -}
{- (Error = 2 is returned).                                     -}
{----------------------------------------------------------------}

var
  PivotRow, Row : integer;
  Dummy : Float;

begin
  { First, find the row with the largest element  }
  PivotRow := ReferenceRow;
  for Row := ReferenceRow + 1 to Dimen do
    if ABS(Coefficients[Row, ReferenceRow]) >
       ABS(Coefficients[PivotRow, ReferenceRow]) then
      PivotRow := Row;
  if PivotRow <> ReferenceRow then
    { Second, switch these two rows  }
    begin
      EROswitch(Coefficients[PivotRow], Coefficients[ReferenceRow]);
      Dummy := Constants[PivotRow];
      Constants[PivotRow] := Constants[ReferenceRow];
      Constants[ReferenceRow] := Dummy;
    end
  else
    { If the diagonal element is zero, no solution exists  }
    if ABS(Coefficients[ReferenceRow, ReferenceRow]) < TNNearlyZero then
      Error := 2;     { No solution  }
end; { procedure Pivot }

begin  { procedure UpperTriangular }
  { Make Coefficients matrix upper triangular  }
  ReferenceRow := 0;
  while (Error = 0) and (ReferenceRow < Dimen - 1) do
  begin
    ReferenceRow := Succ(ReferenceRow);
    { Find row with largest element in this column  }
    { and switch this row with the ReferenceRow     }
    Pivot(Dimen, ReferenceRow, Coefficients, Constants, Error);

    if Error = 0 then
      for Row := ReferenceRow + 1 to Dimen do
        { Make the ReferenceRow element of these rows zero  }
        if ABS(Coefficients[Row, ReferenceRow]) > TNNearlyZero then
        begin
          Multiplier := -Coefficients[Row, ReferenceRow] /
                         Coefficients[ReferenceRow,ReferenceRow];
          EROmultAdd(Multiplier, Dimen,
                     Coefficients[ReferenceRow], Coefficients[Row]);
          Constants[Row] := Constants[Row] +
                            Multiplier * Constants[ReferenceRow];
        end;
  end; { while }
  if ABS(Coefficients[Dimen, Dimen]) < TNNearlyZero then
    Error := 2;    { No solution  }
end; { procedure UpperTriangular }

procedure BackwardsSub(Dimen        : integer;
                   var Coefficients : TNSquareMatrix;
                   var Constants    : TNRowVector;
                   var Solution     : TNRowVector);

{----------------------------------------------------------------}
{- Input: Dimen, Coefficients, Constants                        -}
{- Output: Solution                                             -}
{-                                                              -}
{- This procedure applies backwards substitution to the upper   -}
{- triangular Coefficients matrix and Constants vector. The     -}
{- resulting vector is the solution to the set of equations and -}
{- is stored in the vector Solution.                            -}
{----------------------------------------------------------------}

var
  Term, Row : integer;
  Sum : Float;

begin
  Term := Dimen;
  while Term >= 1 do
  begin
    Sum := 0;
    for Row := Term + 1 to Dimen do
      Sum := Sum + Coefficients[Term, Row] * Solution[Row];
    Solution[Term] := (Constants[Term] - Sum) / Coefficients[Term, Term];
    Term := Pred(Term);
  end;
end; { procedure BackwardsSub }

begin { procedure Partial_Pivoting }
  Initial(Dimen, Coefficients, Constants, Solution, Error);
  if Dimen > 1 then
  begin
    UpperTriangular(Dimen, Coefficients, Constants, Error);
    if Error = 0 then
      BackwardsSub(Dimen, Coefficients, Constants, Solution);
  end;
end; { procedure Partial_Pivoting }

{---------------------------------------------------------------------------}

begin { procedure CreateAndSolveEquations }

  { The following procedure computes Coefficients and    }
  { Constants of the normal equations.  The exact        }
  { solution to the square system of normal equations    }
  { will be the least squares fit to the data.           }
  ComputeNormalEquations(NumPoints, NumTerms, Basis, ZData,
                         Coefficients, Constants);
  Partial_Pivoting(NumTerms, Coefficients, Constants, Solution, Error);
  if Error = 2 then { Returned from Partial_Pivoting  }
    Error := 4; { No solution  }
end; { procedure CreateAndSolveEquations }

procedure TransformSolutionAndFindResiduals(NumPoints         : integer;
                                            NumTerms          : integer;
                                        var YData             : TNColumnVector;
                                        var Solution          : TNRowVector;
                                            Multiplier        : Float;
                                            Constant          : Float;
                                        var Basis             : TNmatrix;
                                        var YFit              : TNColumnVector;
                                        var Residuals         : TNColumnVector;
                                        var StandardDeviation : Float;
                                        var Variance          : Float);

{-------------------------------------------------------------------}
{- Input: NumPoints, NumTerms, YData, Solution, Multiplier,        -}
{-        Constant, Basis                                          -}
{- Output: Solution, YFit, Residuals, StandardDeviation            -}
{-                                                                 -}
{- This procedure computes the goodness of fit of the least        -}
{- squares solution.  The residuals and standard deviation of the  -}
{- fit are returned.  Also, this procedure transforms the solution -}
{- according to the procedure TransformSolution in the include     -}
{- module.  See the particular module for details on the           -}
{- transformation.                                                 -}
{-------------------------------------------------------------------}

procedure ComputeYFitAndResiduals(NumPoints         : integer;
                                  NumTerms          : integer;
                                  Multiplier        : Float;
                                  Constant          : Float;
                              var YData             : TNColumnVector;
                              var Solution          : TNRowVector;
                              var Basis             : TNmatrix;
                              var YFit              : TNColumnVector;
                              var Residuals         : TNColumnVector;
                              var StandardDeviation : Float;
                              var Variance          : Float);

{---------------------------------------------------------------}
{- Input: NumPoints, NumTerms,                                 -}
{- Multiplier, Constant, YData, Solution, Basis                -}
{- Output: YFit, Residuals, StandardDeviation                  -}
{-                                                             -}
{- This procedure computes the value of the least squares      -}
{- approximation at the data points, WData.  The difference    -}
{- between the approximation and the actual values are also    -}
{- computed and are returned in the variable Residuals.  The   -}
{- standard deviation is calculated with the formula:          -}
{-                                                             -}
{-      StandardDeviation = SQRT(Variance)                     -}
{-                                                             -}
{-                                        2                    -}
{-      Variance = SUM(YData[i] - YFit[i]) /                   -}
{-                               (degrees of freedom)          -}
{-                                                             -}
{-      degrees of freedom = NumPoints - NumTerms - 2          -}
{-                                                             -}
{---------------------------------------------------------------}

var
  Index, Term : integer;
  Sum : Float;

begin
  Sum := 0;
  for Index := 1 to NumPoints do
  begin
    YFit[Index] := 0;
    for Term := 1 to NumTerms do
      YFit[Index] := YFit[Index] + Solution[Term]*Basis[Index, Term];

    { The next procedure undoes the transformation of    }
    { the YFit values.  For example, if ZData=Ln(YData)  }
    { then InverseTransform performs the function        }
    { YFit[Index] := Exp(YFit[Index]) so that YFit may   }
    { be compared to YData.                              }
    Fit.InverseTransform(Multiplier, Constant, YFit[Index]);

    Residuals[Index] := YFit[Index] - YData[Index];
    Sum := Sum + Sqr(Residuals[Index]);
  end;
  Variance := Sum;
  if NumPoints = NumTerms then
    StandardDeviation := 0
  else
    StandardDeviation := Sqrt(Sum){!!!}/(NumPoints - NumTerms);
end; { procedure ComputeYFitAndResiduals }

begin  { procedure TransformSolutionAndFindResiduals }
  ComputeYFitAndResiduals(NumPoints, NumTerms, Multiplier, Constant,
                          YData, Solution, Basis, YFit,
                          Residuals, StandardDeviation, Variance);
  Fit.TransformSolution(NumTerms, Solution, Multiplier, Constant, Solution);
end; { procedure TransformSolutionAndFindResiduals }

begin { procedure LeastSquares }

  {added by RRR: initialize LOCAL dynamic arrays!
  because they are zerobased and counted from 1, add extra item
  These arrays are implicitly finalized!
  XData, YData, YFit and Residuals are initialized by caller routine}

  SetLength(WData, NumPoints+1);
  SetLength(ZData, NumPoints+1);
  SetLength(Basis, NumPoints+1);

  InitializeAndFormBasisVectors(NumPoints, XData, YData,
                                Multiplier, Constant, WData, ZData,
                                NumTerms, Solution, Basis, Error);
  if Error = 0  then
    CreateAndSolveEquations(NumPoints, NumTerms, Basis, ZData,
                            Solution, Error);

  if Error = 0 then
    TransformSolutionAndFindResiduals(NumPoints, NumTerms, YData, Solution,
                                      Multiplier, Constant, Basis, YFit,
                                      Residuals, StandardDeviation, Variance);

end; { procedure LeastSquares }

{ TPolyFitFunction }
{-------------------------------------------------------------------------}
{-                                                                       -}
{-            Chebyshev Polynomials                                      -}
{-                                                                       -}
{- This module creates basis vectors to find a least squares solution    -}
{- of the form    f(X) = SUM from i=1 to n of (a[i] * Ti[X]), where Ti   -}
{- is the ith Chebyshev polynomial.  The coefficients of the Ti[X] are   -}
{- converted to coefficients of X^(i-1).                                 -}
{-                                                                       -}
{- The function  ModuleName identifies this as the polynomial fit.       -}
{- The procedure TransformPoly translates and scales the XData to the    -}
{-               interval [-1, 1]. The YData is unchanged.               -}
{- The procedure InverseTransform doesn't do anything in this module.    -}
{- The procedure CreateBasisFunctions creates above basis vectors.       -}
{- The procedure TransformSolution changes the solution vector from      -}
{- coefficients of the Chebyshev polynomials to coefficients of a power  -}
{- series, including adjusting for the shifted data done in TransformPoly-}
{-                                                                       -}
{-------------------------------------------------------------------------}

class procedure TPolyFitFunction.Transform(NumPoints: integer; var XData,
  YData: TNColumnVector; var Multiplier, Constant: Float; var WData,
  ZData: TNColumnVector; var Error: byte);
{---------------------------------------------------------------}
{- Input : NumPoints, XData, YData,                            -}
{- Output: WData, ZData, Error                                 -}
{-                                                             -}
{- This procedure maps the XData linearly into the interval    -}
{- [-1, 1] returning the transformed data in WData.  The YData -}
{- is passed to ZData unchanged.                               -}
{---------------------------------------------------------------}

var
  XDataMin, XDataMax : Float;
  Row : integer;

begin
  XDataMin := XData[1];
  XDataMax := XData[1];
  for Row := 1 to NumPoints do
  begin
    if XDataMin > XData[Row] then
      XDataMin := XData[Row];
    if XDataMax < XData[Row] then
      XDataMax := XData[Row];
  end;
  Multiplier := 2.0 / (XDataMax - XDataMin);
  Constant := - Multiplier * (XDataMax + XDataMin) / 2.0;
  for Row := 1 to NumPoints do
    WData[ Row ] := Multiplier * XData[ Row ] + Constant;
  //// ZDAta := YData;
  for Row := 1 to NumPoints do ZDAta[Row]:=YData[Row];
end; { procedure PolyTransform }

class procedure TPolyFitFunction.InverseTransform(Multiplier,
  Constant: Float; var YFit: Float);
{---------------------------------------------------}
{- Input: Multiplier, Constant, YFit               -}
{- Output: YFit                                    -}
{-                                                 -}
{- This procedure undoes the transformation of     -}
{- the YFit values.  Here, no inverse transform    -}
{- is performed because there was no               -}
{- transformation of Y values in procedure         -}
{- PolyTransform.                                  -}
{---------------------------------------------------}

begin
end; { procedure PolyInverseTransform }

class procedure TPolyFitFunction.CreateBasisFunctions(NumPoints: integer;
  var NumTerms: integer; var WData: TNColumnVector; var Basis: TNmatrix);
{---------------------------------------------------------}
{- Input: NumPoints, NumTerms, WData                     -}
{- Output: Basis                                         -}
{-                                                       -}
{- This procedure creates a matrix of basis vectors.     -}
{- The basis vectors are the CHEBYSHEV POLYNOMIALS.      -}
{-                                                       -}
{- The elements of the matrix are:                       -}
{-     Basis[i, j] = T[j](WData[i])                      -}
{- where T[j](WData[i]) is the jth basis vector          -}
{- evaluated at the value WData[i].                      -}
{-                                                       -}
{- The vectors are:                                      -}
{-   T[1] = 1                                            -}
{-   T[2] = X                                            -}
{-   T[3] = 2x*X - 1                                     -}
{-   T[4] = (4x*X - 3)*X                                 -}
{-   T[5] = (8x*X - 8)*X*X + 1                           -}
{-   etc.                                                -}
{-                                                       -}
{- The Chebyshev Polynomials can be defined recursively: -}
{-   T[1] = 1, T[2] = X                                  -}
{-   T[j] = 2x * T[j - 1] - T[j - 2]                     -}
{-                                                       -}
{---------------------------------------------------------}

var
  Row, Column : integer;

begin
  for Row := 1 to NumPoints do
  begin
    Basis[Row, 1] := 1;
    Basis[Row, 2] := WData[Row];
    for Column := 3 to NumTerms do
      Basis[Row, Column] := 2 * WData[Row] * Basis[Row, Column - 1]
                            - Basis[Row, Column - 2];
  end;
end; { procedure PolyCreateBasisFunctions }

class procedure TPolyFitFunction.TransformSolution(NumTerms: integer;
  var OldSolution: TNRowVector; Multiplier, Constant: Float;
  var NewSolution: TNRowVector);
{--------------------------------------------------------------}
{- Input: NumTerms, OldSolution, Multiplier, Constant         -}
{- Output: NewSolution                                        -}
{-                                                            -}
{- The least squares solution will be more useful if it is    -}
{- expressed as a linear expansion of powers of X, rather     -}
{- than as a linear expansion of Chebyshev polynomials.       -}
{-                                                            -}
{- This procedure converts the coefficients of the Chebyshev  -}
{- polynomials to coefficients of powers of X.                -}
{- The vectors ConversionVec and OldConversionVec store       -}
{- information about the relationship between these two sets  -}
{- of coefficients.  This relationship is defined recursively -}
{- above in the procedure PolyCreateBasisFunctions.           -}
{- The parameters Multiplier and Constant define the linear   -}
{- transformation of the XData, so this is accounted for in   -}
{- finding the polynomial coefficients.                       -}
{--------------------------------------------------------------}

var
  Index, Term : integer;
  Sum : Float;
  OldConversionVec, ConversionVec : TNRowVector;

begin
  FillChar(OldConversionVec, SizeOf(OldConversionVec), 0);
  for Index := 1 to NumTerms do
  begin
    Sum := 0;
    if Index > 1 then
      ConversionVec[Index - 1] := 0;
    for Term := Index to NumTerms do
    begin
      if Term = 1 then
        ConversionVec[Term] := 1.0
      else
        if Term = 2 then
          begin
            if Index = 1 then
              ConversionVec[Term] := Constant
            else
              ConversionVec[Term] := Multiplier
          end
        else
          ConversionVec[Term] := 2 * Multiplier * OldConversionVec[Term - 1]
                                 + 2 * Constant * ConversionVec[Term - 1]
                                 - ConversionVec[Term - 2];
      Sum := Sum + ConversionVec[Term] * OldSolution[Term];
    end;
    NewSolution[Index] := Sum;
    OldConversionVec := ConversionVec;
  end;
end; { procedure PolyTransformSolution }

{-------------------------------------------------------------------------}
{-                                                                       -}
{-            Fourier series                                             -}
{-                                                                       -}
{- This module creates basis vectors to find a least squares solution    -}
{- of the form:  f(x) = a[0] + SUM from i=1 to n/2 of (a[i] - Cos(ix) +  -}
{- a[i+1] - Sin(ix)).  A least squares fit with basis vectors 1, Cos(x), -}
{- Sin(x), Cos(2x), Sin(2x), etc. is made to the data (x, y).            -}
{-                                                                       -}
{- The function ModuleName identifies this as the finite Fourier         -}
{-                              series fit.                              -}
{- The procedure Transform doesn't do anything in this module.           -}
{- The procedure InverseTransform doesn't do anything in this module.    -}
{- The procedure CreateBasisFunctions creates the above basis vectors.   -}
{- The procedure TransformSolution doesn't do anything in this module.   -}
{- The first element of the solution vector will be the constant,        -}
{- the second element will be the coefficient of Cos(x), the third       -}
{- element will be the coefficient of Sin(x), the fourth element will    -}
{- be the coefficient of Cos(2x), etc.                                   -}
{-                                                                       -}
{-------------------------------------------------------------------------}

class procedure TFourierFitFunction.Transform(
                       NumPoints           : integer;
                       var XData           : TNColumnVector;
                       var YData           : TNColumnVector;
                       var Multiplier      : Float;
                       var Constant        : Float;
                       var WData           : TNColumnVector;
                       var ZData           : TNColumnVector;
                       var Error           : byte);

{---------------------------------------------------------------}
{- Input : NumPoints, XData, YData, Multiplier, Constant       -}
{- Output: WData, ZData, Error                                 -}
{-                                                             -}
{- No transformations are needed for Fourier Series            -}
{---------------------------------------------------------------}

var i: integer;      // WARNING!!! dynamic arrays cannot be assigned normally!
begin
  for i:=1 to NumPoints do
  begin
    WData[i]:=XData[i];
    ZData[i]:=YData[i];
  end;
end; { procedure FourierTransform }

class procedure TFourierFitFunction.InverseTransform(
                            Multiplier : Float;
                            Constant   : Float;
                            var YFit   : Float);

{---------------------------------------------------}
{- Input: Multiplier, Constant, YFit               -}
{- Output: YFit                                    -}
{-                                                 -}
{- This procedure undoes the transformation of     -}
{- the YFit values.  Here, no inverse transform    -}
{- is performed because there was no               -}
{- transformation in procedure Transform.          -}
{---------------------------------------------------}

begin
end; { procedure FourierInverseTransform }

class procedure TFourierFitFunction.CreateBasisFunctions(
                                      NumPoints : integer;
                                  var NumTerms  : integer;
                                  var WData     : TNColumnVector;
                                  var Basis     : TNmatrix);

{-------------------------------------------------------------}
{- Input: NumPoints, NumTerms, WData                         -}
{- Output: Basis                                             -}
{-                                                           -}
{- This procedure creates a matrix of basis vectors.         -}
{- The basis vectors are the FOURIER SERIES.                 -}
{-                                                           -}
{- The elements of the matrix are:                           -}
{-     Basis[i, j] = F[j](WData[i])                          -}
{- where F[j](WData[i]) is the jth basis vector              -}
{- evaluated at the value WData[i].                          -}
{-                                                           -}
{- The vectors are:                                          -}
{-      F[1] = 1                                             -}
{-      F[2] = Cos(x);                                       -}
{-      F[3] = Sin(x);                                       -}
{-      F[4] = Cos(2x);                                      -}
{-      F[5] = Sin(2x);                                      -}
{-      F[6] = Cos(3x);                                      -}
{-      F[7] = Sin(3x);                                      -}
{-      etc.                                                 -}
{-                                                           -}
{- This series is defined recursively by:                    -}
{-      F[1] = 1, F[2] = Cos(x),  F[3] = Sin(x)              -}
{-      F[j] = F[2] - F[j - 2] - F[3] - F[j - 1]  for even j -}
{-      F[j] = F[3] - F[j - 3] + F[2] - F[j - 2]  for odd j  -}
{-------------------------------------------------------------}

var
  Row, Column : integer;

begin
  for Row := 1 to NumPoints do
  begin
    Basis[Row, 1] := 1;
    Basis[Row, 2] := Cos(WData[Row]);
    Basis[Row, 3] := Sin(WData[Row]);
    for Column := 4 to NumTerms do
      if Odd(Column) then
        Basis[Row, Column] := Basis[Row, 3] * Basis[Row, Column-3]
                            + Basis[Row, 2] * Basis[Row, Column-2]
      else
        Basis[Row, Column] := Basis[Row, 2] * Basis[Row, Column-2]
                            - Basis[Row, 3] * Basis[Row, Column-1];
  end;
end; { procedure FourierCreateBasisFunctions }

class procedure TFourierFitFunction.TransformSolution(
                                   NumTerms        : integer;
                               var OldSolution     : TNRowVector;
                                   Multiplier      : Float;
                                   Constant        : Float;
                               var NewSolution     : TNRowVector);

{----------------------------------------------------}
{- Input: NumTerms, OldSolution, DummyMultiplier,   -}
{-        DummyConstant                             -}
{- Output: NewSolution                              -}
{-                                                  -}
{- No need to change the coefficients of the        -}
{- Fourier series.                                  -}
{----------------------------------------------------}

begin
  NewSolution := OldSolution    { no transformation  }
end; { procedure FourierTransformSolution }

{----------------------------------------------------------------------------}
{-                                                                          -}
{-            Y = AX^B                                                      -}
{-                                                                          -}
{- This module creates basis vectors to find a least squares solution       -}
{- of the form    f(X) = A * X^B.  Taking the logarithm of both sides:      -}
{- Ln(f(X)) = Ln(A) + B * Ln(X).  A linear least squares fit                -}
{- (i.e. with basis vectors Ln(X) and 1) is then made to the data           -}
{- (Ln(X), Ln(Y)). The slope will be B, and the intercept will be Ln(A).    -}
{-                                                                          -}
{- The function ModuleName identifies this as the power law fit.            -}
{- The procedure Transform converts the data from (X, Y) to (Ln(X), Ln(Y)). -}
{- The procedure InverseTransform converts from YFit to Exp(YFit)           -}
{- The procedure CreateBasisFunctions creates the vectors Ln(X) and 1.      -}
{- The procedure TransformSolution changes the solution vector from         -}
{- (Ln(A), B) to (A, B).  Therefore, the first coefficient is A,            -}
{- and the second coefficient is B.                                         -}
{-                                                                          -}
{----------------------------------------------------------------------------}

class procedure TPowerFitFunction.Transform(NumPoints: integer;
                     var XData         : TNColumnVector;
                     var YData         : TNColumnVector;
                     var Multiplier    : Float;
                     var Constant      : Float;
                     var WData         : TNColumnVector;
                     var ZData         : TNColumnVector;
                     var Error         : byte);

{---------------------------------------------------------------}
{- Input : NumPoints, XData, YData, Multiplier, Constant       -}
{- Output: WData, ZData, Error                                 -}
{-                                                             -}
{- This procedure transforms the X and Y values to their       -}
{- logarithms.  A linear least squares fit will then be made   -}
{- to to Ln(Y) = B * Ln(X) + Ln(A). If the Y values are of     -}
{- differing sign, Error = 3 is returned.                      -}
{---------------------------------------------------------------}

var
  Index : integer;
  YPoint : Float;

begin
  Index := 0;
  if YData[1] < 0 then
    Multiplier := -1
  else
    Multiplier := 1;
  while (Index < NumPoints) and (Error = 0) do
  begin
    Index := Succ(Index);
    if XData[Index] <= 0 then
      Error := 3
    else
      begin
        YPoint := Multiplier * YData[Index];
        if YPoint <= 0 then { The data must all have the same sign  }
          Error := 3
        else
          begin
            WData[Index] := Ln(XData[Index]);
            ZData[Index] := Ln(YPoint);
          end;
      end;
  end; { while }
end; { procedure PowerTransform }

class procedure TPowerFitFunction.InverseTransform(Multiplier    : Float;
                                                   Constant      : Float;
                                               var YFit          : Float);

{-------------------------------------------------}
{- Input: Multiplier, DummyConstant, YFit        -}
{- Output: YFit                                  -}
{-                                               -}
{- This procedure undoes the transformation of   -}
{- the YFit values.  Here, the function          -}
{- YFit := Exp(YFit) is performed to undo the    -}
{- Ln transformation in procedure Transform.     -}
{-------------------------------------------------}

begin
  YFit := Multiplier * Exp(YFit);
end; { procedure PowerInverseTransform }

class procedure TPowerFitFunction.CreateBasisFunctions(NumPoints : integer;
                                var NumTerms  : integer;
                                var WData     : TNColumnVector;
                                var Basis     : TNmatrix);

{---------------------------------------------------------}
{- Input: NumPoints, NumTerms, WData                     -}
{- Output: Basis                                         -}
{-                                                       -}
{- This procedure creates a matrix of basis vectors.     -}
{- The elements of the matrix are:                       -}
{-     Basis[i, j] = C[j](WData[i])                      -}
{- where C[j](WData[i]) is the jth basis vector          -}
{- evaluated at the value WData[i].                      -}
{-                                                       -}
{-                                                       -}
{- The vectors are:                                      -}
{-   C[1] = 1                                            -}
{-   C[2] = X                                            -}
{---------------------------------------------------------}

var
  Row : integer;

begin
  NumTerms := 2;  { This is only a straight line least squares  }
  for Row := 1 to NumPoints do
  begin
    Basis[Row, 1] := 1;
    Basis[Row, 2] := WData[Row];
  end;
end; { procedure PowerCreateBasisFunctions }

class procedure TPowerFitFunction.TransformSolution(NumTerms: integer;
                             var OldSolution   : TNRowVector;
                                 Multiplier    : Float;
                                 Constant      : Float;
                             var NewSolution   : TNRowVector);

{--------------------------------------------------------------}
{- Input: NumTerms, OldSolution, Multiplier, DummyConstant    -}
{- Output: NewSolution                                        -}
{-                                                            -}
{- The least squares solution will be more useful if it is    -}
{- expressed in terms of Y = AX^B, rather than in terms       -}
{- of Ln(Y) = B * Ln(X) + Ln(A).                              -}
{--------------------------------------------------------------}

begin
  NewSolution[1] := Multiplier * Exp(OldSolution[1]);
end; { procedure PowerTransformSolution }

{-------------------------------------------------------------------------}
{-                                                                       -}
{-            Y = A * Exp(bx)                                            -}
{-                                                                       -}
{- This module creates basis vectors to find a least squares solution    -}
{- of the form    f(X) = A * Exp(bx).  Taking the logarithm of both      -}
{- sides:   Ln(f(X)) = Ln(A) + B * X. A linear least squares fit         -}
{- (i.e. with basis vectors X and 1) is then made to the data            -}
{- (X, Ln(Y)). The slope will be B, and the intercept will be Ln(A).     -}
{-                                                                       -}
{- The function ModuleName identifies this as the exponential fit        -}
{- The procedure Transform converts the data from (X, Y) to (X, Ln(Y)).  -}
{- The procedure InverseTransform converts from YFit to Exp(YFit)        -}
{- The procedure CreateBasisFunctions creates the vectors 1 and X.       -}
{- The procedure TransformSolution changes the solution vector from      -}
{- (Ln(A), B) to (A, B).  Therefore, the first coefficient is a,         -}
{- and the second coefficient is B.                                      -}
{-                                                                       -}
{-------------------------------------------------------------------------}

class procedure TExpoFitFunction.Transform(NumPoints: integer;
                    var XData         : TNColumnVector;
                    var YData         : TNColumnVector;
                    var Multiplier    : Float;
                    var Constant      : Float;
                    var WData         : TNColumnVector;
                    var ZData         : TNColumnVector;
                    var Error         : byte);

{---------------------------------------------------------------}
{- Input : NumPoints, XData, YData, Multiplier, DummyConstant  -}
{- Output: WData, ZData, Error                                 -}
{-                                                             -}
{- This procedure transforms the Y values to their             -}
{- logarithms.  A linear least squares fit will then be made   -}
{- to Ln(Y) = bx + Ln(A). If the Y values are of different     -}
{- sign, then Error = 3 is returned.                           -}
{---------------------------------------------------------------}

var
  Index : integer;
  YPoint : Float;

begin
  //  WData := XData;
  for Index:=1 to NumPoints do WData[Index]:=XData[Index]; //////////
  
  if  YData[1] < 0 then
    Multiplier := -1
  else
    Multiplier := 1;
  Index := 0;
  while (Index < NumPoints) and (Error = 0) do
  begin
    Index := Succ(Index);
    YPoint := Multiplier * YData[Index];
    if YPoint <= 0 then
      Error := 3   { The Y values must all have the same sign }
    else
      ZData[Index] := Ln(YPoint);
  end;
end; { procedure ExpoTransform }

class procedure TExpoFitFunction.InverseTransform(Multiplier: Float;
Constant: Float; var YFit: Float);

{-------------------------------------------------}
{- Input: Multiplier, DummyConstant, YFit        -}
{- Output: YFit                                  -}
{-                                               -}
{- This procedure undoes the transformation of   -}
{- the YFit values.  Here, the function          -}
{- YFit := Exp(YFit) is performed to undo the    -}
{- Ln transformation in procedure Transform.     -}
{-------------------------------------------------}

begin
  YFit := Multiplier * Exp(YFit);
end; { procedure ExpoInverseTransform }

class procedure TExpoFitFunction.CreateBasisFunctions(NumPoints : integer;
                               var NumTerms  : integer;
                               var WData     : TNColumnVector;
                               var Basis     : TNmatrix);

{---------------------------------------------------------}
{- Input: NumPoints, NumTerms, WData                     -}
{- Output: Basis                                         -}
{-                                                       -}
{- This procedure creates a matrix of basis vectors.     -}
{-                                                       -}
{- The elements of the matrix are:                       -}
{-     Basis[i, j] = C[j](WData[i])                      -}
{- where C[j](WData[i]) is the jth basis vector          -}
{- evaluated at the value WData[i].                      -}
{-                                                       -}
{- The vectors are:                                      -}
{-   C[1] = 1                                            -}
{-   C[2] = X                                            -}
{---------------------------------------------------------}

var
  Row : integer;

begin
  NumTerms := 2;  { This is only a straight line least squares }
  for Row := 1 to NumPoints do
  begin
    Basis[Row, 1] := 1;
    Basis[Row, 2] := WData[Row];
  end;
end; { procedure ExpoCreateBasisFunctions }

class procedure TExpoFitFunction.TransformSolution(NumTerms: integer;
                            var OldSolution   : TNRowVector;
                                Multiplier    : Float;
                                Constant      : Float;
                            var NewSolution   : TNRowVector);

{--------------------------------------------------------------}
{- Input: NumTerms, OldSolution, Multiplier, DummyConstant    -}
{- Output: NewSolution                                        -}
{-                                                            -}
{- The least squares solution will be more useful if it is    -}
{- expressed in terms of Y = A - Exp(bx), rather than in      -}
{- terms of Ln(Y) = B - X + Ln(A).                            -}
{--------------------------------------------------------------}

begin
  NewSolution[1] := Multiplier * Exp(OldSolution[1]);
end; { procedure ExpoTransformSolution }

{ TLorentzFitFunction }

// Lorentzian: Y=A*X/((X-Xo)^2+B*X) NOTE: without additive constant!
// Transformation Y'=X/Y gives polynom: Y'=X^2/A+X*(B-2Xo)/A+Xo^2/A
// P1'=A=1/P3
// P2'=Xo=sqrt(P1/P3)
// P3'=B=P2/P3+2*sqrt(P1/P3)

class procedure TLorentzFitFunction.CreateBasisFunctions(NumPoints: integer;
  var NumTerms: integer; var WData: TNColumnVector; var Basis: TNmatrix);
begin
  NumTerms:=3; inherited;
end;

class procedure TLorentzFitFunction.Transform(NumPoints: integer; var XData,
  YData: TNColumnVector; var Multiplier, Constant: Float; var WData,
  ZData: TNColumnVector; var Error: byte);
var I: integer;
begin
  for I:=1 to NumPoints do if YData[I]<>0 then YDAta[I]:=XData[I]/YData[I]
  else begin Error:=5; Break; end;
  inherited;
end;

class procedure TLorentzFitFunction.TransformSolution(NumTerms: integer;
  var OldSolution: TNRowVector; Multiplier, Constant: Float;
  var NewSolution: TNRowVector);
var a,b,x0: Float;
begin
  inherited;
  a:=1/NewSolution[3];
  x0:=sqrt(abs(NewSolution[1]/NewSolution[3]));  // ABS! this may cause errors
  b:=NewSolution[2]/NewSolution[3]+2*x0;
  NewSolution[1]:=a; NewSolution[2]:=x0; NewSolution[3]:=b;
end;

{ TSineFitFunction }

class procedure TSineFitFunction.CreateBasisFunctions(NumPoints: integer;
  var NumTerms: integer; var WData: TNColumnVector; var Basis: TNmatrix);
begin
  NumTerms:=3; inherited;
end;

class procedure TSineFitFunction.TransformSolution(NumTerms: integer;
  var OldSolution: TNRowVector; Multiplier, Constant: Float;
  var NewSolution: TNRowVector);
var Amp,Phase: float;
begin
  inherited;
  Amp:=sqrt(sqr(NewSolution[2])+sqr(NewSolution[3]));
  if NewSolution[3]=0 then Phase:=Pi/2
  else Phase:=ArcTan(NewSolution[2]/NewSolution[3]);
  NewSolution[2]:=Amp; NewSolution[3]:=Phase;
end;

end.

